/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.noties.markwon

import android.content.Context
import android.widget.TextView
import io.noties.markwon.Markwon.TextSetter
import io.noties.markwon.MarkwonVisitorFactory.Companion.create
import io.noties.markwon.core.MarkwonTheme.builderWithDefaults
import java.util.Collections
import org.commonmark.parser.Parser

/** @since 3.0.0 */
internal class MarkwonBuilderImpl(private val context: Context) : Markwon.Builder {
    private val plugins: MutableList<MarkwonPlugin> = ArrayList(3)
    private var bufferType = TextView.BufferType.SPANNABLE
    private var textSetter: TextSetter? = null

    // @since 4.4.0
    private var fallbackToRawInputWhenEmpty = true

    override fun bufferType(bufferType: TextView.BufferType): Markwon.Builder {
        this.bufferType = bufferType
        return this
    }

    override fun textSetter(textSetter: TextSetter): Markwon.Builder {
        this.textSetter = textSetter
        return this
    }

    override fun usePlugin(plugin: MarkwonPlugin): Markwon.Builder {
        plugins.add(plugin)
        return this
    }

    override fun usePlugins(plugins: Iterable<MarkwonPlugin?>): Markwon.Builder {
        val iterator = plugins.iterator()
        var plugin: MarkwonPlugin
        while (iterator.hasNext()) {
            plugin = iterator.next()!!
            this.plugins.add(plugin)
        }
        return this
    }

    override fun fallbackToRawInputWhenEmpty(
        fallbackToRawInputWhenEmpty: Boolean
    ): Markwon.Builder {
        this.fallbackToRawInputWhenEmpty = fallbackToRawInputWhenEmpty
        return this
    }

    override fun build(): Markwon {
        check(plugins.isNotEmpty()) {
            "No plugins were added to this builder. Use #usePlugin " + "method to add them"
        }

        // please note that this method must not modify supplied collection
        // if nothing should be done -> the same collection can be returned
        val plugins = preparePlugins(plugins)
        val parserBuilder = Parser.Builder()
        val themeBuilder = builderWithDefaults(context)
        val configurationBuilder = MarkwonConfiguration.Builder()
        val visitorBuilder: MarkwonVisitor.Builder = MarkwonVisitorImpl.BuilderImpl()
        val spanFactoryBuilder: MarkwonSpansFactory.Builder = MarkwonSpansFactoryImpl.BuilderImpl()
        for (plugin in plugins) {
            plugin.configureParser(parserBuilder)
            plugin.configureTheme(themeBuilder)
            plugin.configureConfiguration(configurationBuilder)
            plugin.configureVisitor(visitorBuilder)
            plugin.configureSpansFactory(spanFactoryBuilder)
        }
        val configuration =
            configurationBuilder.build(themeBuilder.build(), spanFactoryBuilder.build())

        // @since 4.1.1
        // @since 4.1.2 - do not reuse render-props (each render call should have own render-props)
        val visitorFactory = create(visitorBuilder, configuration)
        return MarkwonImpl(
            bufferType,
            textSetter,
            parserBuilder.build(),
            visitorFactory,
            configuration,
            Collections.unmodifiableList(plugins),
            fallbackToRawInputWhenEmpty)
    }

    companion object {
        private fun preparePlugins(plugins: List<MarkwonPlugin>): List<MarkwonPlugin> {
            return RegistryImpl(plugins).process()
        }
    }
}
