/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.noties.markwon

import java.util.Collections
import org.commonmark.node.Node

/** @since 3.0.0 */
internal class MarkwonSpansFactoryImpl(private val factories: Map<Class<out Node>, SpanFactory>) :
    MarkwonSpansFactory {
    override fun <N : Node?> get(node: Class<N>): SpanFactory? {
        return factories[node]
    }

    override fun <N : Node?> require(node: Class<N>): SpanFactory {
        return get(node) ?: throw NullPointerException(node.name)
    }

    internal class BuilderImpl : MarkwonSpansFactory.Builder {
        private val factories: MutableMap<Class<out Node>, SpanFactory> = HashMap(3)

        override fun <N : Node> setFactory(
            node: Class<N>,
            factory: SpanFactory
        ): MarkwonSpansFactory.Builder {
            factories[node] = factory
            return this
        }

        @Deprecated("", ReplaceWith("prependFactory(node, factory)"))
        override fun <N : Node> addFactory(
            node: Class<N>,
            factory: SpanFactory
        ): MarkwonSpansFactory.Builder {
            return prependFactory(node, factory)
        }

        override fun <N : Node> appendFactory(
            node: Class<N>,
            factory: SpanFactory
        ): MarkwonSpansFactory.Builder {
            val existing = factories[node]
            if (existing == null) {
                factories[node] = factory
            } else {
                if (existing is CompositeSpanFactory) {
                    existing.factories.add(0, factory)
                } else {
                    val compositeSpanFactory = CompositeSpanFactory(factory, existing)
                    factories[node] = compositeSpanFactory
                }
            }
            return this
        }

        override fun <N : Node> prependFactory(
            node: Class<N>,
            factory: SpanFactory
        ): MarkwonSpansFactory.Builder {
            // if there is no factory registered for this node -> just add it
            val existing = factories[node]
            if (existing == null) {
                factories[node] = factory
            } else {
                // existing span factory can be of CompositeSpanFactory at this point -> append to
                // it
                if (existing is CompositeSpanFactory) {
                    existing.factories.add(factory)
                } else {
                    // if it's not composite at this point -> make it
                    val compositeSpanFactory = CompositeSpanFactory(existing, factory)
                    factories[node] = compositeSpanFactory
                }
            }
            return this
        }

        override fun <N : Node> getFactory(node: Class<N>): SpanFactory? {
            return factories[node]
        }

        override fun <N : Node> requireFactory(node: Class<N>): SpanFactory {
            return getFactory(node) ?: throw NullPointerException(node.name)
        }

        override fun build(): MarkwonSpansFactory {
            return MarkwonSpansFactoryImpl(Collections.unmodifiableMap(factories))
        }
    }

    internal class CompositeSpanFactory(first: SpanFactory, second: SpanFactory) : SpanFactory {
        val factories: MutableList<SpanFactory>

        init {
            factories = ArrayList(3)
            factories.add(first)
            factories.add(second)
        }

        override fun getSpans(configuration: MarkwonConfiguration, props: RenderProps): Any {
            // please note that we do not check it factory itself returns an array of spans,
            // as this behaviour is supported now (previously we supported only a single-level
            // array)
            val length = factories.size
            val out = arrayOfNulls<Any>(length)
            for (i in 0 until length) {
                out[i] = factories[i].getSpans(configuration, props)
            }
            return out
        }
    }
}
