/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.noties.markwon

import android.os.Build
import android.text.Spanned
import android.util.Log
import android.widget.TextView
import androidx.core.text.PrecomputedTextCompat
import io.noties.markwon.Markwon.TextSetter
import java.lang.ref.WeakReference
import java.util.concurrent.Executor

/**
 * Please note this class requires `androidx.core:core` artifact being explicitly added to your
 * dependencies. Please do not use with `markwon-recycler` as it will lead to bad item rendering
 * (due to async nature)
 *
 * @see io.noties.markwon.Markwon.TextSetter
 * @since 4.1.0
 */
class PrecomputedTextSetterCompat internal constructor(private val executor: Executor) :
    TextSetter {
    override fun setText(
        textView: TextView,
        markdown: Spanned,
        bufferType: TextView.BufferType,
        onComplete: Runnable
    ) {
        val reference = WeakReference(textView)
        executor.execute {
            try {
                val precomputedTextCompat = precomputedText(reference.get(), markdown)
                if (precomputedTextCompat != null) {
                    applyText(reference.get(), precomputedTextCompat, bufferType, onComplete)
                }
            } catch (t: Throwable) {
                if (BuildConfig.DEBUG)
                    Log.e("PrecomputdTxtSetterCmpt", "Exception during pre-computing text", t)
                // apply initial markdown
                applyText(reference.get(), markdown, bufferType, onComplete)
            }
        }
    }

    companion object {
        /** @param executor for background execution of text pre-computation */
        fun create(executor: Executor): PrecomputedTextSetterCompat {
            return PrecomputedTextSetterCompat(executor)
        }

        private fun precomputedText(textView: TextView?, spanned: Spanned): PrecomputedTextCompat? {
            if (textView == null) {
                return null
            }
            val params: PrecomputedTextCompat.Params =
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.P) {
                    // use native parameters on P
                    PrecomputedTextCompat.Params(textView.textMetricsParams)
                } else {
                    val builder = PrecomputedTextCompat.Params.Builder(textView.paint)

                    // please note that text-direction initialization is omitted
                    // by default it will be determined by the first locale-specific character
                    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                        // another miss on API surface, this can easily be done by the compat class
                        // itself
                        builder
                            .setBreakStrategy(textView.breakStrategy)
                            .setHyphenationFrequency(textView.hyphenationFrequency)
                    }
                    builder.build()
                }
            return PrecomputedTextCompat.create(spanned, params)
        }

        private fun applyText(
            textView: TextView?,
            text: Spanned,
            bufferType: TextView.BufferType,
            onComplete: Runnable
        ) {
            textView?.post {
                textView.setText(text, bufferType)
                onComplete.run()
            }
        }
    }
}
