/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.noties.markwon;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import io.noties.markwon.core.CorePlugin;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

// @since 4.0.0
class RegistryImpl implements MarkwonPlugin.Registry {

  private final List<MarkwonPlugin> origin;
  private final List<MarkwonPlugin> plugins;
  private final Set<MarkwonPlugin> pending;

  RegistryImpl(@NonNull List<MarkwonPlugin> origin) {
    this.origin = origin;
    this.plugins = new ArrayList<>(origin.size());
    this.pending = new HashSet<>(3);
  }

  @Nullable
  private static <P extends MarkwonPlugin> P find(
      @NonNull List<MarkwonPlugin> plugins, @NonNull Class<P> type) {
    for (MarkwonPlugin plugin : plugins) {
      if (type.isAssignableFrom(plugin.getClass())) {
        //noinspection unchecked
        return (P) plugin;
      }
    }
    return null;
  }

  @NonNull
  @Override
  public <P extends MarkwonPlugin> P require(@NonNull Class<P> plugin) {
    return get(plugin);
  }

  @Override
  public <P extends MarkwonPlugin> void require(
      @NonNull Class<P> plugin, @NonNull MarkwonPlugin.Action<? super P> action) {
    action.apply(get(plugin));
  }

  @NonNull
  List<MarkwonPlugin> process() {
    for (MarkwonPlugin plugin : origin) {
      configure(plugin);
    }
    return plugins;
  }

  private void configure(@NonNull MarkwonPlugin plugin) {

    // important -> check if it's in plugins
    //  if it is -> no need to configure (already configured)

    if (!plugins.contains(plugin)) {

      if (pending.contains(plugin)) {
        throw new IllegalStateException("Cyclic dependency chain found: " + pending);
      }

      // start tracking plugins that are pending for configuration
      pending.add(plugin);

      plugin.configure(this);

      // stop pending tracking
      pending.remove(plugin);

      // check again if it's included (a child might've configured it already)
      // add to out-collection if not already present
      // this is a bit different from `find` method as it does check for exact instance
      // and not a sub-type
      if (!plugins.contains(plugin)) {
        // core-plugin must always be the first one (if it's present)
        if (CorePlugin.class.isAssignableFrom(plugin.getClass())) {
          plugins.add(0, plugin);
        } else {
          plugins.add(plugin);
        }
      }
    }
  }

  @NonNull
  private <P extends MarkwonPlugin> P get(@NonNull Class<P> type) {

    // check if present already in plugins
    // find in origin, if not found -> throw, else add to out-plugins

    P plugin = find(plugins, type);

    if (plugin == null) {

      plugin = find(origin, type);

      if (plugin == null) {
        throw new IllegalStateException(
            "Requested plugin is not added: " + type.getName() + ", plugins: " + origin);
      }

      configure(plugin);
    }

    return plugin;
  }
}
