/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.noties.markwon.ext.tables

import android.text.Spanned
import io.noties.markwon.Markwon
import org.commonmark.ext.gfm.tables.TableBlock
import org.commonmark.ext.gfm.tables.TableCell
import org.commonmark.ext.gfm.tables.TableHead
import org.commonmark.ext.gfm.tables.TableRow
import org.commonmark.node.AbstractVisitor
import org.commonmark.node.CustomNode

/**
 * A class to parse `TableBlock` and return a data-structure that is not dependent on
 * commonmark-java table extension. Can be useful when rendering tables require special handling
 * (multiple views, specific table view) for example when used with `markwon-recycler` artifact
 *
 * @see .parse
 * @since 3.0.0
 */
class Table(private val rows: List<Row>) {
    fun rows(): List<Row> {
        return rows
    }

    override fun toString(): String {
        return "Table{rows=$rows}"
    }

    enum class Alignment {
        LEFT,
        CENTER,
        RIGHT
    }

    class Row(private val isHeader: Boolean, private val columns: List<Column>) {
        fun header(): Boolean {
            return isHeader
        }

        fun columns(): List<Column> {
            return columns
        }

        override fun toString(): String {
            return "Row{isHeader=$isHeader, columns=$columns}"
        }
    }

    class Column(private val alignment: Alignment, private val content: Spanned) {
        fun alignment(): Alignment {
            return alignment
        }

        fun content(): Spanned {
            return content
        }

        override fun toString(): String {
            return "Column{alignment=$alignment, content=$content}"
        }
    }

    internal class ParseVisitor(private val markwon: Markwon) : AbstractVisitor() {
        private var rows: MutableList<Row>? = null
        private var pendingRow: MutableList<Column>? = null
        private var pendingRowIsHeader = false

        fun rows(): List<Row>? {
            return rows
        }

        override fun visit(customNode: CustomNode) {
            if (customNode is TableCell) {
                if (pendingRow == null) {
                    pendingRow = ArrayList(2)
                }
                pendingRow!!.add(
                    Column(alignment(customNode.alignment), markwon.render(customNode)))
                pendingRowIsHeader = customNode.isHeader
                return
            }
            if (customNode is TableHead || customNode is TableRow) {
                visitChildren(customNode)

                // this can happen, ignore such row
                if (pendingRow != null && pendingRow!!.size > 0) {
                    if (rows == null) {
                        rows = ArrayList(2)
                    }
                    rows!!.add(Row(pendingRowIsHeader, pendingRow!!))
                }
                pendingRow = null
                pendingRowIsHeader = false
                return
            }
            visitChildren(customNode)
        }

        companion object {
            private fun alignment(alignment: TableCell.Alignment): Alignment {
                val out: Alignment =
                    if (TableCell.Alignment.RIGHT == alignment) {
                        Alignment.RIGHT
                    } else if (TableCell.Alignment.CENTER == alignment) {
                        Alignment.CENTER
                    } else {
                        Alignment.LEFT
                    }
                return out
            }
        }
    }

    companion object {
        /**
         * Factory method to obtain an instance of [Table]
         *
         * @param markwon Markwon
         * @param tableBlock TableBlock to parse
         * @return parsed [Table] or null
         */
        fun parse(markwon: Markwon, tableBlock: TableBlock): Table? {
            val table: Table?
            val visitor = ParseVisitor(markwon)
            tableBlock.accept(visitor)
            val rows = visitor.rows()
            table =
                if (rows == null) {
                    null
                } else {
                    Table(rows)
                }
            return table
        }
    }
}
