/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.noties.markwon.ext.tasklist

import android.content.Context
import android.graphics.drawable.Drawable
import android.util.TypedValue
import androidx.annotation.AttrRes
import androidx.annotation.ColorInt
import io.noties.markwon.AbstractMarkwonPlugin
import io.noties.markwon.MarkwonSpansFactory
import io.noties.markwon.MarkwonVisitor
import io.noties.markwon.R
import org.commonmark.parser.Parser

/** @since 3.0.0 */
class TaskListPlugin private constructor(private val drawable: Drawable) : AbstractMarkwonPlugin() {
    override fun configureParser(builder: Parser.Builder) {
        builder.postProcessor(TaskListPostProcessor())
    }

    override fun configureSpansFactory(builder: MarkwonSpansFactory.Builder) {
        builder.setFactory(TaskListItem::class.java, TaskListSpanFactory(drawable))
    }

    override fun configureVisitor(builder: MarkwonVisitor.Builder) {
        builder.on(TaskListItem::class.java) { visitor, taskListItem ->
            val length = visitor.length()
            visitor.visitChildren(taskListItem)
            TaskListProps.DONE[visitor.renderProps()] = taskListItem.isDone
            visitor.setSpansForNode(taskListItem, length)
            if (visitor.hasNext(taskListItem)) {
                visitor.ensureNewLine()
            }
        }
    }

    companion object {
        /**
         * Supplied Drawable must be stateful ([Drawable.isStateful] returns true). If a task is
         * marked as done, then this drawable will be updated with an `int[] {
         * android.R.attr.state_checked }` as the state, otherwise an empty array will be used. This
         * library provides a ready to be used Drawable: [TaskListDrawable]
         *
         * @see TaskListDrawable
         */
        fun create(drawable: Drawable): TaskListPlugin {
            return TaskListPlugin(drawable)
        }

        fun create(context: Context): TaskListPlugin {

            // by default we will be using link color for the checkbox color
            // & window background as a checkMark color
            val linkColor = resolve(context, R.attr.textColorLink)
            val backgroundColor = resolve(context, R.attr.colorBackground)
            return TaskListPlugin(TaskListDrawable(linkColor, linkColor, backgroundColor))
        }

        fun create(
            @ColorInt checkedFillColor: Int,
            @ColorInt normalOutlineColor: Int,
            @ColorInt checkMarkColor: Int
        ): TaskListPlugin {
            return TaskListPlugin(
                TaskListDrawable(checkedFillColor, normalOutlineColor, checkMarkColor))
        }

        private fun resolve(context: Context, @AttrRes attr: Int): Int {
            val typedValue = TypedValue()
            val attrs = intArrayOf(attr)
            val typedArray = context.obtainStyledAttributes(typedValue.data, attrs)
            return try {
                typedArray.getColor(0, 0)
            } finally {
                typedArray.recycle()
            }
        }
    }
}
