/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.noties.markwon.html

import java.util.Collections

internal abstract class HtmlTagImpl
protected constructor(
    @JvmField val name: String,
    @JvmField val start: Int,
    val attributes: Map<String, String>
) : HtmlTag {
    @JvmField var end = HtmlTag.NO_END

    override fun name(): String {
        return name
    }

    override fun start(): Int {
        return start
    }

    override fun end(): Int {
        return end
    }

    override fun isEmpty(): Boolean {
        return start == end
    }

    override fun attributes(): Map<String, String> {
        return attributes
    }

    override fun isClosed(): Boolean {
        return end > HtmlTag.NO_END
    }

    abstract fun closeAt(end: Int)

    internal class InlineImpl(name: String, start: Int, attributes: Map<String, String>) :
        HtmlTagImpl(name, start, attributes), HtmlTag.Inline {
        override fun closeAt(end: Int) {
            if (!isClosed) {
                super.end = end
            }
        }

        override fun toString(): String {
            return "InlineImpl{" +
                "name='" +
                name +
                '\'' +
                ", start=" +
                start +
                ", end=" +
                end +
                ", attributes=" +
                attributes +
                '}'
        }

        override fun isInline(): Boolean {
            return true
        }

        override fun isBlock(): Boolean {
            return false
        }

        override fun getAsInline(): HtmlTag.Inline {
            return this
        }

        override fun getAsBlock(): HtmlTag.Block {
            throw ClassCastException("Cannot cast Inline instance to Block")
        }
    }

    internal class BlockImpl(
        name: String,
        start: Int,
        attributes: Map<String, String>,
        @JvmField val parent: BlockImpl?
    ) : HtmlTagImpl(name, start, attributes), HtmlTag.Block {
        @JvmField var children: List<BlockImpl>? = null

        override fun closeAt(end: Int) {
            if (!isClosed) {
                super.end = end
                if (children != null) {
                    for (child in children!!) {
                        child.closeAt(end)
                    }
                }
            }
        }

        override fun isRoot(): Boolean {
            return parent == null
        }

        override fun parent(): HtmlTag.Block? {
            return parent
        }

        override fun children(): List<HtmlTag.Block> {
            val list: List<HtmlTag.Block> =
                if (children == null) {
                    emptyList()
                } else {
                    Collections.unmodifiableList(children as List<HtmlTag.Block>?)
                }
            return list
        }

        override fun attributes(): Map<String, String> {
            return attributes
        }

        override fun isInline(): Boolean {
            return false
        }

        override fun isBlock(): Boolean {
            return true
        }

        override fun getAsInline(): HtmlTag.Inline {
            throw ClassCastException("Cannot cast Block instance to Inline")
        }

        override fun getAsBlock(): HtmlTag.Block {
            return this
        }

        override fun toString(): String {
            return "BlockImpl{" +
                "name='" +
                name +
                '\'' +
                ", start=" +
                start +
                ", end=" +
                end +
                ", attributes=" +
                attributes +
                ", parent=" +
                parent?.name +
                ", children=" +
                children +
                '}'
        }

        companion object {
            @JvmStatic
            fun root(): BlockImpl {
                return BlockImpl("", 0, emptyMap(), null)
            }

            fun create(
                name: String,
                start: Int,
                attributes: Map<String, String>,
                parent: BlockImpl?
            ): BlockImpl {
                return BlockImpl(name, start, attributes, parent)
            }
        }
    }
}
