/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.noties.markwon.html.jsoup.nodes;

import io.noties.markwon.html.jsoup.helper.Validate;
import java.util.Map;
import java.util.Objects;

/** A single key + value attribute. (Only used for presentation.) */
public class Attribute implements Map.Entry<String, String>, Cloneable {

  Attributes
      parent; // used to update the holding Attributes when the key / value is changed via this
  // interface
  private String key;
  private String val;

  /**
   * Create a new attribute from unencoded (raw) key and value.
   *
   * @param key attribute key; case is preserved.
   * @param value attribute value
   */
  public Attribute(String key, String value) {
    this(key, value, null);
  }

  /**
   * Create a new attribute from unencoded (raw) key and value.
   *
   * @param key attribute key; case is preserved.
   * @param val attribute value
   * @param parent the containing Attributes (this Attribute is not automatically added to said
   *     Attributes)
   */
  public Attribute(String key, String val, Attributes parent) {
    Validate.notNull(key);
    this.key = key.trim();
    Validate.notEmpty(key); // trimming could potentially make empty, so validate here
    this.val = val;
    this.parent = parent;
  }

  /**
   * Get the attribute key.
   *
   * @return the attribute key
   */
  public String getKey() {
    return key;
  }

  /**
   * Set the attribute key; case is preserved.
   *
   * @param key the new key; must not be null
   */
  public void setKey(String key) {
    Validate.notNull(key);
    key = key.trim();
    Validate.notEmpty(key); // trimming could potentially make empty, so validate here
    if (parent != null) {
      int i = parent.indexOfKey(this.key);
      if (i != Attributes.NotFound) parent.keys[i] = key;
    }
    this.key = key;
  }

  /**
   * Get the attribute value.
   *
   * @return the attribute value
   */
  public String getValue() {
    return val;
  }

  /**
   * Set the attribute value.
   *
   * @param val the new attribute value; must not be null
   */
  public String setValue(String val) {
    String oldVal = parent.get(this.key);
    if (parent != null) {
      int i = parent.indexOfKey(this.key);
      if (i != Attributes.NotFound) parent.vals[i] = val;
    }
    this.val = val;
    return oldVal;
  }

  @Override
  public boolean equals(Object o) { // note parent not considered
    if (this == o) return true;
    if (o == null || getClass() != o.getClass()) return false;
    Attribute attribute = (Attribute) o;
    if (!Objects.equals(key, attribute.key)) return false;
    return Objects.equals(val, attribute.val);
  }

  @Override
  public int hashCode() { // note parent not considered
    int result = key != null ? key.hashCode() : 0;
    result = 31 * result + (val != null ? val.hashCode() : 0);
    return result;
  }

  @Override
  public Attribute clone() {
    try {
      return (Attribute) super.clone();
    } catch (CloneNotSupportedException e) {
      throw new RuntimeException(e);
    }
  }
}
