/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.noties.markwon.image.coil

import android.content.Context
import android.graphics.drawable.Drawable
import android.text.Spanned
import android.widget.TextView
import coil.Coil.imageLoader
import coil.ImageLoader
import coil.request.Disposable
import coil.request.ImageRequest
import io.noties.markwon.AbstractMarkwonPlugin
import io.noties.markwon.MarkwonConfiguration
import io.noties.markwon.MarkwonSpansFactory
import io.noties.markwon.image.AsyncDrawable
import io.noties.markwon.image.AsyncDrawableLoader
import io.noties.markwon.image.AsyncDrawableScheduler
import io.noties.markwon.image.DrawableUtils.applyIntrinsicBoundsIfEmpty
import io.noties.markwon.image.ImageSpanFactory
import java.util.concurrent.atomic.AtomicBoolean
import org.commonmark.node.Image

/**
 * @author Tyler Wong
 * @since 4.2.0
 */
class CoilImagesPlugin internal constructor(coilStore: CoilStore, imageLoader: ImageLoader) :
    AbstractMarkwonPlugin() {
    interface CoilStore {
        fun load(drawable: AsyncDrawable): ImageRequest

        fun cancel(disposable: Disposable)
    }

    private val coilAsyncDrawableLoader: CoilAsyncDrawableLoader

    init {
        coilAsyncDrawableLoader = CoilAsyncDrawableLoader(coilStore, imageLoader)
    }

    override fun configureSpansFactory(builder: MarkwonSpansFactory.Builder) {
        builder.setFactory(Image::class.java, ImageSpanFactory())
    }

    override fun configureConfiguration(builder: MarkwonConfiguration.Builder) {
        builder.asyncDrawableLoader(coilAsyncDrawableLoader)
    }

    override fun beforeSetText(textView: TextView, markdown: Spanned) {
        AsyncDrawableScheduler.unschedule(textView)
    }

    override fun afterSetText(textView: TextView) {
        AsyncDrawableScheduler.schedule(textView)
    }

    private class CoilAsyncDrawableLoader(coilStore: CoilStore, imageLoader: ImageLoader) :
        AsyncDrawableLoader() {
        private val coilStore: CoilStore
        private val imageLoader: ImageLoader
        private val cache: HashMap<AsyncDrawable, Disposable> = HashMap(2)

        init {
            this.coilStore = coilStore
            this.imageLoader = imageLoader
        }

        override fun load(drawable: AsyncDrawable) {
            val loaded = AtomicBoolean(false)
            val target: coil.target.Target = AsyncDrawableTarget(drawable, loaded)
            val request: ImageRequest = coilStore.load(drawable).newBuilder().target(target).build()
            // @since 4.5.1 execute can return result _before_ disposable is created,
            //  thus `execute` would finish before we put disposable in cache (and thus result is
            //  not delivered)
            val disposable: Disposable = imageLoader.enqueue(request)
            // if flag was not set, then job is running (else - finished before we got here)
            if (!loaded.get()) {
                // mark flag
                loaded.set(true)
                cache[drawable] = disposable
            }
        }

        override fun cancel(drawable: AsyncDrawable) {
            val disposable: Disposable? = cache.remove(drawable)
            if (disposable != null) {
                coilStore.cancel(disposable)
            }
        }

        override fun placeholder(drawable: AsyncDrawable): Drawable? {
            return null
        }

        private inner class AsyncDrawableTarget(
            private val drawable: AsyncDrawable,
            private val loaded: AtomicBoolean
        ) : coil.target.Target {
            override fun onSuccess(result: Drawable) {
                // @since 4.5.1 check finished flag (result can be delivered _before_ disposable is
                // created)
                if (cache.remove(drawable) != null || !loaded.get()) {
                    // mark
                    loaded.set(true)
                    if (drawable.isAttached) {
                        applyIntrinsicBoundsIfEmpty(result)
                        drawable.setResult(result)
                    }
                }
            }

            override fun onStart(placeholder: Drawable?) {
                if (placeholder != null && drawable.isAttached) {
                    applyIntrinsicBoundsIfEmpty(placeholder)
                    drawable.setResult(placeholder)
                }
            }

            override fun onError(error: Drawable?) {
                if (cache.remove(drawable) != null) {
                    if (error != null && drawable.isAttached) {
                        applyIntrinsicBoundsIfEmpty(error)
                        drawable.setResult(error)
                    }
                }
            }
        }
    }

    companion object {
        fun create(context: Context): CoilImagesPlugin {
            return create(
                object : CoilStore {
                    override fun load(drawable: AsyncDrawable): ImageRequest {
                        return ImageRequest.Builder(context).data(drawable.destination).build()
                    }

                    override fun cancel(disposable: Disposable) {
                        disposable.dispose()
                    }
                },
                imageLoader(context))
        }

        fun create(context: Context, imageLoader: ImageLoader): CoilImagesPlugin {
            return create(
                object : CoilStore {
                    override fun load(drawable: AsyncDrawable): ImageRequest {
                        return ImageRequest.Builder(context).data(drawable.destination).build()
                    }

                    override fun cancel(disposable: Disposable) {
                        disposable.dispose()
                    }
                },
                imageLoader)
        }

        fun create(coilStore: CoilStore, imageLoader: ImageLoader): CoilImagesPlugin {
            return CoilImagesPlugin(coilStore, imageLoader)
        }
    }
}
